import numpy as np


def move_left_upper_2d_point_to_zero(x):
    x_trans = np.amin(x[:, 0])
    y_trans = np.amin(x[:, 1])
    x[:, 0] -= x_trans
    x[:, 1] -= y_trans
    return x, x_trans, y_trans









import os, sys
import shutil
import matplotlib.pyplot as plt
import numpy as np
import torch
import matplotlib

def get_rgba_from_cmap(cmap, val):
    cmap = matplotlib.cm.get_cmap(cmap)
    return np.asarray(cmap(val)).reshape(1,-1)

def get_device(gpu_id=-1):
    if gpu_id < 0:
        return torch.device('cpu')
    else:
        return torch.device(f'cuda:{gpu_id}')

def get_coloset_elem_dist(row, arr):
    assert len(arr.shape) == 2
    row = np.asarray(row)
    dist = np.sum((arr - row) ** 2, axis=1, keepdims=True)
    idx_min = dist.argmin(axis=0)[0]
    return dist[idx_min], idx_min

def get_row_index_in_arr(row, arr):
    assert len(arr.shape) == 2
    row = np.asarray(row)
    min_dist, min_dist_idx = get_coloset_elem_dist(row, arr)
    return min_dist_idx if min_dist < 1e-6 else False


def make_path_if_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)


def remove_make_path(path):
    if os.path.exists(path):
        shutil.rmtree(path)
    os.makedirs(path)


def copy_move_file(source, target):
    shutil.copy(source, target)

def draw_confusion_matrix(y_pred, y_true):
    # https: // vitalflux.com / accuracy - precision - recall - f1 - score - python - example /
    from sklearn.metrics import confusion_matrix
    from sklearn.metrics import precision_score, recall_score, f1_score, accuracy_score

    conf_matrix = confusion_matrix(y_true=y_true, y_pred=y_pred)
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.matshow(conf_matrix, cmap=plt.cm.Oranges, alpha=0.3)
    for i in range(conf_matrix.shape[0]):
        for j in range(conf_matrix.shape[1]):
            ax.text(x=j, y=i, s=conf_matrix[i, j], va='center', ha='center', size='xx-large')

    plt.xlabel('Predictions', fontsize=18)
    plt.ylabel('Actuals', fontsize=18)
    plt.title('Confusion Matrix', fontsize=18)
    plt.savefig('confusion_matrix.png')
    print('Num. of Obj.: %.3d' % len(y_true))
    print('Num. Pos./ Num. Neg.: {}/{}'.format(y_true.count(1), y_true.count(0)))
    print('Precision: %.3f' % precision_score(y_true, y_pred))
    print('Recall: %.3f' % recall_score(y_true, y_pred))
    print('Accuracy: %.3f' % accuracy_score(y_true, y_pred))
    print('F1 Score: %.3f' % f1_score(y_true, y_pred))




import random
import time
import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import gym
import numpy as np
import torch

from tianshou.data import (
    Batch,
    CachedReplayBuffer,
    ReplayBuffer,
    ReplayBufferManager,
    VectorReplayBuffer,
    to_numpy,
    Collector
)
from tianshou.data.batch import _alloc_by_keys_diff
from tianshou.env import BaseVectorEnv, DummyVectorEnv
from tianshou.policy import BasePolicy

T = 100


def get_new_trajactory_len(buffers, env_id, ep_len, relabelling_strategy, q_net):
    env_buffer = buffers[env_id]
    env_buffer_len = env_buffer.last_index[0] + 1
    traj_len = ep_len[env_id]
    obs_index_range = np.arange(env_buffer_len - traj_len, env_buffer_len) % len(env_buffer)
    original_trajectory = env_buffer[obs_index_range]

    hindsight_goal_id = original_trajectory.obs_next.hindsight_goal_id
    hindsight_goal_id = np.unique(hindsight_goal_id[hindsight_goal_id >= 0])
    max_rew = original_trajectory.obs_next.max_rew[0]

    if len(hindsight_goal_id) == 0:
        return None, None

    if relabelling_strategy == 'advantage':
        s0_id = original_trajectory.obs.state_id[0]
        s0_id = [s0_id] * len(hindsight_goal_id)
        obs = Batch({'state_id': s0_id, 'goal_id': hindsight_goal_id})
        with torch.no_grad():
            q, _ = q_net(obs, state=None, info=None)
            q = torch.amax(q, dim=-1).cpu().numpy()

    new_trajactory_len = []
    for index, hindsight_goal_id_i in enumerate(hindsight_goal_id):
        length_candidate = np.where(original_trajectory.obs_next.hindsight_goal_id == hindsight_goal_id_i)[0]
        if relabelling_strategy == 'min_length':
            length = np.amin(length_candidate)
        elif relabelling_strategy == 'random':
            length = random.choice(length_candidate)
        elif relabelling_strategy == 'advantage':
            length = np.amin(length_candidate)
            q_i = q[index]
            cumulative_reward = -(length - 1) * (1 / T) + max_rew
            if q_i > cumulative_reward:
                continue
        else:
            raise NotImplementedError
        new_trajactory_len.append(length)

    if len(new_trajactory_len) == 0:
        return None, None

    return new_trajactory_len, original_trajectory


# class Collector(object):
#     """Collector enables the policy to interact with different types of envs with \
#     exact number of steps or episodes.
#
#     :param policy: an instance of the :class:`~tianshou.policy.BasePolicy` class.
#     :param env: a ``gym.Env`` environment or an instance of the
#         :class:`~tianshou.env.BaseVectorEnv` class.
#     :param buffer: an instance of the :class:`~tianshou.data.ReplayBuffer` class.
#         If set to None, it will not store the data. Default to None.
#     :param function preprocess_fn: a function called before the data has been added to
#         the buffer, see issue #42 and :ref:`preprocess_fn`. Default to None.
#     :param bool exploration_noise: determine whether the action needs to be modified
#         with corresponding policy's exploration noise. If so, "policy.
#         exploration_noise(act, batch)" will be called automatically to add the
#         exploration noise into action. Default to False.
#
#     The "preprocess_fn" is a function called before the data has been added to the
#     buffer with batch format. It will receive only "obs" and "env_id" when the
#     collector resets the environment, and will receive six keys "obs_next", "rew",
#     "done", "info", "policy" and "env_id" in a normal env step. It returns either a
#     dict or a :class:`~tianshou.data.Batch` with the modified keys and values. Examples
#     are in "test/base/test_collector.py".
#
#     .. note::
#
#         Please make sure the given environment has a time limitation if using n_episode
#         collect option.
#     """
#
#     def __init__(
#             self,
#             policy: BasePolicy,
#             env: Union[gym.Env, BaseVectorEnv],
#             buffer: Optional[ReplayBuffer] = None,
#             preprocess_fn: Optional[Callable[..., Batch]] = None,
#             exploration_noise: bool = False,
#     ) -> None:
#         super().__init__()
#         if isinstance(env, gym.Env) and not hasattr(env, "__len__"):
#             warnings.warn("Single environment detected, wrap to DummyVectorEnv.")
#             env = DummyVectorEnv([lambda: env])
#         self.env = env
#         self.env_num = len(env)
#         self.exploration_noise = exploration_noise
#         self._assign_buffer(buffer)
#         self.policy = policy
#         self.preprocess_fn = preprocess_fn
#         self._action_space = env.action_space
#         # avoid creating attribute outside __init__
#         self.reset()
#
#     def _assign_buffer(self, buffer: Optional[ReplayBuffer]) -> None:
#         """Check if the buffer matches the constraint."""
#         if buffer is None:
#             buffer = VectorReplayBuffer(self.env_num, self.env_num)
#         elif isinstance(buffer, ReplayBufferManager):
#             assert buffer.buffer_num >= self.env_num
#             if isinstance(buffer, CachedReplayBuffer):
#                 assert buffer.cached_buffer_num >= self.env_num
#         else:  # ReplayBuffer or PrioritizedReplayBuffer
#             assert buffer.maxsize > 0
#             if self.env_num > 1:
#                 if type(buffer) == ReplayBuffer:
#                     buffer_type = "ReplayBuffer"
#                     vector_type = "VectorReplayBuffer"
#                 else:
#                     buffer_type = "PrioritizedReplayBuffer"
#                     vector_type = "PrioritizedVectorReplayBuffer"
#                 raise TypeError(
#                     f"Cannot use {buffer_type}(size={buffer.maxsize}, ...) to collect "
#                     f"{self.env_num} envs,\n\tplease use {vector_type}(total_size="
#                     f"{buffer.maxsize}, buffer_num={self.env_num}, ...) instead."
#                 )
#         self.buffer = buffer
#
#     def reset(self) -> None:
#         """Reset all related variables in the collector."""
#         # use empty Batch for "state" so that self.data supports slicing
#         # convert empty Batch to None when passing data to policy
#         self.data = Batch(
#             obs={}, act={}, rew={}, done={}, obs_next={}, info={}, policy={}
#         )
#         self.reset_env()
#         self.reset_buffer()
#         self.reset_stat()
#
#     def reset_stat(self) -> None:
#         """Reset the statistic variables."""
#         self.collect_step, self.collect_episode, self.collect_time = 0, 0, 0.0
#
#     def reset_buffer(self, keep_statistics: bool = False) -> None:
#         """Reset the data buffer."""
#         self.buffer.reset(keep_statistics=keep_statistics)
#
#     def reset_env(self) -> None:
#         """Reset all of the environments."""
#         obs = self.env.reset()
#         if self.preprocess_fn:
#             obs = self.preprocess_fn(obs=obs,
#                                      env_id=np.arange(self.env_num)).get("obs", obs)
#         self.data.obs = obs
#
#     def _reset_state(self, id: Union[int, List[int]]) -> None:
#         """Reset the hidden state: self.data.state[id]."""
#         if hasattr(self.data.policy, "hidden_state"):
#             state = self.data.policy.hidden_state  # it is a reference
#             if isinstance(state, torch.Tensor):
#                 state[id].zero_()
#             elif isinstance(state, np.ndarray):
#                 state[id] = None if state.dtype == object else 0
#             elif isinstance(state, Batch):
#                 state.empty_(id)
#
#     def collect(
#             self,
#             n_step: Optional[int] = None,
#             n_episode: Optional[int] = None,
#             random: bool = False,
#             render: Optional[float] = None,
#             no_grad: bool = True,
#     ) -> Dict[str, Any]:
#         """Collect a specified number of step or episode.
#
#         To ensure unbiased sampling result with n_episode option, this function will
#         first collect ``n_episode - env_num`` episodes, then for the last ``env_num``
#         episodes, they will be collected evenly from each env.
#
#         :param int n_step: how many steps you want to collect.
#         :param int n_episode: how many episodes you want to collect.
#         :param bool random: whether to use random policy for collecting data. Default
#             to False.
#         :param float render: the sleep time between rendering consecutive frames.
#             Default to None (no rendering).
#         :param bool no_grad: whether to retain gradient in policy.forward(). Default to
#             True (no gradient retaining).
#
#         .. note::
#
#             One and only one collection number specification is permitted, either
#             ``n_step`` or ``n_episode``.
#
#         :return: A dict including the following keys
#
#             * ``n/ep`` collected number of episodes.
#             * ``n/st`` collected number of steps.
#             * ``rews`` array of episode reward over collected episodes.
#             * ``lens`` array of episode length over collected episodes.
#             * ``idxs`` array of episode start index in buffer over collected episodes.
#             * ``rew`` mean of episodic rewards.
#             * ``len`` mean of episodic lengths.
#             * ``rew_std`` standard error of episodic rewards.
#             * ``len_std`` standard error of episodic lengths.
#         """
#         assert not self.env.is_async, "Please use AsyncCollector if using async venv."
#         if n_step is not None:
#             assert n_episode is None, (
#                 f"Only one of n_step or n_episode is allowed in Collector."
#                 f"collect, got n_step={n_step}, n_episode={n_episode}."
#             )
#             assert n_step > 0
#             if not n_step % self.env_num == 0:
#                 warnings.warn(
#                     f"n_step={n_step} is not a multiple of #env ({self.env_num}), "
#                     "which may cause extra transitions collected into the buffer."
#                 )
#             ready_env_ids = np.arange(self.env_num)
#         elif n_episode is not None:
#             assert n_episode > 0
#             ready_env_ids = np.arange(min(self.env_num, n_episode))
#             self.data = self.data[:min(self.env_num, n_episode)]
#         else:
#             raise TypeError(
#                 "Please specify at least one (either n_step or n_episode) "
#                 "in AsyncCollector.collect()."
#             )
#
#         start_time = time.time()
#
#         step_count = 0
#         episode_count = 0
#         episode_rews = []
#         episode_lens = []
#         episode_start_indices = []
#
#         while True:
#             assert len(self.data) == len(ready_env_ids)
#             # restore the state: if the last state is None, it won't store
#             last_state = self.data.policy.pop("hidden_state", None)
#
#             # get the next action
#             if random:
#                 self.data.update(
#                     act=[self._action_space[i].sample() for i in ready_env_ids]
#                 )
#             else:
#                 if no_grad:
#                     with torch.no_grad():  # faster than retain_grad version
#                         # self.data.obs will be used by agent to get result
#                         result = self.policy(self.data, last_state)
#                 else:
#                     result = self.policy(self.data, last_state)
#                 # update state / act / policy into self.data
#                 policy = result.get("policy", Batch())
#                 assert isinstance(policy, Batch)
#                 state = result.get("state", None)
#                 if state is not None:
#                     policy.hidden_state = state  # save state into buffer
#                 act = to_numpy(result.act)
#                 if self.exploration_noise:
#                     act = self.policy.exploration_noise(act, self.data)
#                 self.data.update(policy=policy, act=act)
#
#             # get bounded and remapped actions first (not saved into buffer)
#             action_remap = self.policy.map_action(self.data.act)
#             # step in env
#             result = self.env.step(action_remap, ready_env_ids)  # type: ignore
#             obs_next, rew, done, info = result
#
#             self.data.update(obs_next=obs_next, rew=rew, done=done, info=info)
#             if self.preprocess_fn:
#                 self.data.update(
#                     self.preprocess_fn(
#                         obs_next=self.data.obs_next,
#                         rew=self.data.rew,
#                         done=self.data.done,
#                         info=self.data.info,
#                         policy=self.data.policy,
#                         env_id=ready_env_ids,
#                     )
#                 )
#
#             if render:
#                 self.env.render()
#                 if render > 0 and not np.isclose(render, 0):
#                     time.sleep(render)
#
#             # add data into the buffer
#             ptr, ep_rew, ep_len, ep_idx = self.buffer.add(
#                 self.data, buffer_ids=ready_env_ids
#             )
#
#             # collect statistics
#             step_count += len(ready_env_ids)
#
#             if np.any(done):
#                 env_ind_local = np.where(done)[0]
#                 env_ind_global = ready_env_ids[env_ind_local]
#                 episode_count += len(env_ind_local)
#                 episode_lens.append(ep_len[env_ind_local])
#                 episode_rews.append(ep_rew[env_ind_local])
#                 episode_start_indices.append(ep_idx[env_ind_local])
#                 # now we copy obs_next to obs, but since there might be
#                 # finished episodes, we have to reset finished envs first.
#                 obs_reset = self.env.reset(env_ind_global)
#                 if self.preprocess_fn:
#                     obs_reset = self.preprocess_fn(
#                         obs=obs_reset, env_id=env_ind_global
#                     ).get("obs", obs_reset)
#                 self.data.obs_next[env_ind_local] = obs_reset
#                 for i in env_ind_local:
#                     self._reset_state(i)
#
#                 # remove surplus env id from ready_env_ids
#                 # to avoid bias in selecting environments
#                 if n_episode:
#                     surplus_env_num = len(ready_env_ids) - (n_episode - episode_count)
#                     if surplus_env_num > 0:
#                         mask = np.ones_like(ready_env_ids, dtype=bool)
#                         mask[env_ind_local[:surplus_env_num]] = False
#                         ready_env_ids = ready_env_ids[mask]
#                         self.data = self.data[mask]
#
#             self.data.obs = self.data.obs_next
#
#             if (n_step and step_count >= n_step) or \
#                     (n_episode and episode_count >= n_episode):
#                 break
#
#         # generate statistics
#         self.collect_step += step_count
#         self.collect_episode += episode_count
#         self.collect_time += max(time.time() - start_time, 1e-9)
#
#         if n_episode:
#             self.data = Batch(
#                 obs={}, act={}, rew={}, done={}, obs_next={}, info={}, policy={}
#             )
#             self.reset_env()
#
#         if episode_count > 0:
#             rews, lens, idxs = list(
#                 map(
#                     np.concatenate,
#                     [episode_rews, episode_lens, episode_start_indices]
#                 )
#             )
#             rew_mean, rew_std = rews.mean(), rews.std()
#             len_mean, len_std = lens.mean(), lens.std()
#         else:
#             rews, lens, idxs = np.array([]), np.array([], int), np.array([], int)
#             rew_mean = rew_std = len_mean = len_std = 0
#
#         return {
#             "n/ep": episode_count,
#             "n/st": step_count,
#             "rews": rews,
#             "lens": lens,
#             "idxs": idxs,
#             "rew": rew_mean,
#             "len": len_mean,
#             "rew_std": rew_std,
#             "len_std": len_std,
#         }

class HERCollector(Collector):
    def __init__(self,
                 policy: BasePolicy,
                 env: Union[gym.Env, BaseVectorEnv],
                 buffer: Optional[ReplayBuffer] = None,
                 preprocess_fn: Optional[Callable[..., Batch]] = None,
                 exploration_noise: bool = False,
                 ):
        super(HERCollector, self).__init__(
            policy,
            env,
            buffer,
            preprocess_fn,
            exploration_noise
        )

        self.relabelling_strategy = 'random'

    def collect(
            self,
            n_step: Optional[int] = None,
            n_episode: Optional[int] = None,
            random: bool = False,
            render: Optional[float] = None,
            no_grad: bool = True,
    ) -> Dict[str, Any]:
        """Collect a specified number of step or episode.

        To ensure unbiased sampling result with n_episode option, this function will
        first collect ``n_episode - env_num`` episodes, then for the last ``env_num``
        episodes, they will be collected evenly from each env.

        :param int n_step: how many steps you want to collect.
        :param int n_episode: how many episodes you want to collect.
        :param bool random: whether to use random policy for collecting data. Default
            to False.
        :param float render: the sleep time between rendering consecutive frames.
            Default to None (no rendering).
        :param bool no_grad: whether to retain gradient in policy.forward(). Default to
            True (no gradient retaining).

        .. note::

            One and only one collection number specification is permitted, either
            ``n_step`` or ``n_episode``.

        :return: A dict including the following keys

            * ``n/ep`` collected number of episodes.
            * ``n/st`` collected number of steps.
            * ``rews`` array of episode reward over collected episodes.
            * ``lens`` array of episode length over collected episodes.
            * ``idxs`` array of episode start index in buffer over collected episodes.
            * ``rew`` mean of episodic rewards.
            * ``len`` mean of episodic lengths.
            * ``rew_std`` standard error of episodic rewards.
            * ``len_std`` standard error of episodic lengths.
        """
        assert not self.env.is_async, "Please use AsyncCollector if using async venv."
        if n_step is not None:
            assert n_episode is None, (
                f"Only one of n_step or n_episode is allowed in Collector."
                f"collect, got n_step={n_step}, n_episode={n_episode}."
            )
            assert n_step > 0
            if not n_step % self.env_num == 0:
                warnings.warn(
                    f"n_step={n_step} is not a multiple of #env ({self.env_num}), "
                    "which may cause extra transitions collected into the buffer."
                )
            ready_env_ids = np.arange(self.env_num)
        elif n_episode is not None:
            assert n_episode > 0
            ready_env_ids = np.arange(min(self.env_num, n_episode))
            self.data = self.data[:min(self.env_num, n_episode)]
        else:
            raise TypeError(
                "Please specify at least one (either n_step or n_episode) "
                "in AsyncCollector.collect()."
            )

        start_time = time.time()

        step_count = 0
        episode_count = 0
        episode_rews = []
        episode_lens = []
        episode_start_indices = []

        while True:
            assert len(self.data) == len(ready_env_ids)
            # restore the state: if the last state is None, it won't store
            last_state = self.data.policy.pop("hidden_state", None)

            # get the next action
            if random:
                self.data.update(
                    act=[self._action_space[i].sample() for i in ready_env_ids]
                )
            else:
                if no_grad:
                    with torch.no_grad():  # faster than retain_grad version
                        # self.data.obs will be used by agent to get result
                        result = self.policy(self.data, last_state)
                else:
                    result = self.policy(self.data, last_state)
                # update state / act / policy into self.data
                policy = result.get("policy", Batch())
                assert isinstance(policy, Batch)
                state = result.get("state", None)
                if state is not None:
                    policy.hidden_state = state  # save state into buffer
                act = to_numpy(result.act)
                if self.exploration_noise:
                    act = self.policy.exploration_noise(act, self.data)
                self.data.update(policy=policy, act=act)

            # get bounded and remapped actions first (not saved into buffer)
            action_remap = self.policy.map_action(self.data.act)
            # step in env
            result = self.env.step(action_remap, ready_env_ids)  # type: ignore
            obs_next, rew, done, info = result

            self.data.update(obs_next=obs_next, rew=rew, done=done, info=info)
            if self.preprocess_fn:
                self.data.update(
                    self.preprocess_fn(
                        obs_next=self.data.obs_next,
                        rew=self.data.rew,
                        done=self.data.done,
                        info=self.data.info,
                        policy=self.data.policy,
                        env_id=ready_env_ids,
                    )
                )

            if render:
                self.env.render()
                if render > 0 and not np.isclose(render, 0):
                    time.sleep(render)

            # add data into the buffer
            ptr, ep_rew, ep_len, ep_idx = self.buffer.add(
                self.data, buffer_ids=ready_env_ids
            )

            # collect statistics
            step_count += len(ready_env_ids)

            if np.any(done):
                env_ind_local = np.where(done)[0]
                env_ind_global = ready_env_ids[env_ind_local]
                episode_count += len(env_ind_local)
                episode_lens.append(ep_len[env_ind_local])
                episode_rews.append(ep_rew[env_ind_local])
                episode_start_indices.append(ep_idx[env_ind_local])
                # now we copy obs_next to obs, but since there might be
                # finished episodes, we have to reset finished envs first.
                obs_reset = self.env.reset(env_ind_global)
                if self.preprocess_fn:
                    obs_reset = self.preprocess_fn(
                        obs=obs_reset, env_id=env_ind_global
                    ).get("obs", obs_reset)
                self.data.obs_next[env_ind_local] = obs_reset
                for i in env_ind_local:
                    self._reset_state(i)

                # remove surplus env id from ready_env_ids
                # to avoid bias in selecting environments
                if n_episode:
                    surplus_env_num = len(ready_env_ids) - (n_episode - episode_count)
                    if surplus_env_num > 0:
                        mask = np.ones_like(ready_env_ids, dtype=bool)
                        mask[env_ind_local[:surplus_env_num]] = False
                        ready_env_ids = ready_env_ids[mask]
                        self.data = self.data[mask]

                # use HER to create more trajectory
                for env_id in env_ind_global:  # enumerate env
                    # get recently collected data from buffer

                    new_trajactory_len, original_trajectory = get_new_trajactory_len(
                        buffers=self.buffer.buffers, env_id=env_id, ep_len=ep_len,
                        relabelling_strategy=self.relabelling_strategy,
                        q_net=self.policy.model,
                    )

                    if new_trajactory_len is None:
                        continue

                    for length in new_trajactory_len:  # use different mid goal to resample
                        trajectory = Batch(original_trajectory[:length + 1], copy=True)

                        new_goal = trajectory.obs_next.hindsight_goal_id[length]

                        new_goals = np.repeat([new_goal], length + 1, axis=0)
                        trajectory.obs.goal_id = new_goals
                        trajectory.obs_next.goal_id = new_goals
                        trajectory.rew[-1] = trajectory.obs_next.max_rew[0]
                        trajectory.done[-1] = True
                        for transition in trajectory:
                            self.buffer.buffers[env_id].add(transition)

                # ---------------------------------------------
            self.data.obs = self.data.obs_next

            if (n_step and step_count >= n_step) or \
                    (n_episode and episode_count >= n_episode):
                break

        # generate statistics
        self.collect_step += step_count
        self.collect_episode += episode_count
        self.collect_time += max(time.time() - start_time, 1e-9)

        if n_episode:
            self.data = Batch(
                obs={}, act={}, rew={}, done={}, obs_next={}, info={}, policy={}
            )
            self.reset_env()

        if episode_count > 0:
            rews, lens, idxs = list(
                map(
                    np.concatenate,
                    [episode_rews, episode_lens, episode_start_indices]
                )
            )
            rew_mean, rew_std = rews.mean(), rews.std()
            len_mean, len_std = lens.mean(), lens.std()
        else:
            rews, lens, idxs = np.array([]), np.array([], int), np.array([], int)
            rew_mean = rew_std = len_mean = len_std = 0

        return {
            "n/ep": episode_count,
            "n/st": step_count,
            "rews": rews,
            "lens": lens,
            "idxs": idxs,
            "rew": rew_mean,
            "len": len_mean,
            "rew_std": rew_std,
            "len_std": len_std,
        }

class GHRLCollector(HERCollector):
    def __init__(self,
                 policy: BasePolicy,
                 env: Union[gym.Env, BaseVectorEnv],
                 buffer: Optional[ReplayBuffer] = None,
                 preprocess_fn: Optional[Callable[..., Batch]] = None,
                 exploration_noise: bool = False,
                 ):
        super(GHRLCollector, self).__init__(
            policy,
            env,
            buffer,
            preprocess_fn,
            exploration_noise
        )

        self.relabelling_strategy = 'advantage'



import sys
import os

sys.path.append(os.path.dirname(__file__))
import numpy as np
import cv2
import ctypes

from tempfile import NamedTemporaryFile
import matplotlib.pyplot as plt
import matplotlib
from get_foreground_mask import remove


# from .get_foreground_mask import remove


class LineSegmentDetector():
    def __init__(self,
                 scale=0.8,
                 sigma_scale=0.6,
                 ang_th=22.5,
                 quant=2.0,
                 eps=0.0,
                 density_th=0.7,
                 n_bins=1024,
                 max_grad=255.0
                 ):
        """
        :param scale: Scale the image by Gaussian filter to 'scale'.
        :param sigma_scale: Sigma for Gaussian filter is computed as sigma = sigma_scale/scale.
        :param ang_th: Gradient angle tolerance in degrees.
        :param quant: Bound to the quantization error on the gradient norm.
        :param eps: Detection threshold, -log10(NFA).
        :param density_th: Minimal density of region points in rectangle.
        :param n_bins: Number of bins in pseudo-ordering of gradient modulus.
        :param max_grad: Gradient modulus in the highest bin. The default value corresponds to
                the highest gradient modulus on images with gray levels in [0,255].
        """
        self.lsdlib = self.load_lsd_library()
        if self.lsdlib is None:
            raise ImportError('Cannot load dynamic library. Did you compile LSD?')

        self.scale = scale
        self.sigma_scale = sigma_scale
        self.ang_th = ang_th
        self.quant = quant
        self.eps = eps
        self.density_th = density_th
        self.n_bins = n_bins
        self.max_grad = max_grad

    def load_lsd_library(self, ):
        root_dir = os.path.abspath(os.path.dirname(__file__))

        libnames = ['linux/liblsd.so']
        libdir = os.path.join(os.path.dirname(__file__), 'lib')  # 'lib'

        if sys.platform == 'win32':
            if sys.maxsize > 2 ** 32:
                libnames = ['win32/x64/lsd.dll', 'win32/x64/liblsd.dll']
            else:
                libnames = ['win32/x86/lsd.dll', 'win32/x86/liblsd.dll']

        elif sys.platform == 'darwin':
            libnames = ['darwin/liblsd.dylib']

        while root_dir is not None:
            for libname in libnames:
                try:
                    lsdlib = ctypes.cdll[os.path.join(root_dir, libdir, libname)]
                    return lsdlib
                except Exception as e:
                    pass
            tmp = os.path.dirname(root_dir)
            if tmp == root_dir:
                root_dir = None
            else:
                root_dir = tmp

        # if we didn't find the library so far, try loading without
        # a full path as a last resort
        for libname in libnames:
            try:
                lsdlib = ctypes.cdll[libname]
                return lsdlib
            except Exception as e:
                pass

        return None

    def visualize_line_segment(self, image, lines, mode='overlay'):
        mask = 255 * np.ones(image.shape[:2])
        for i in range(lines.shape[0]):
            pt1 = (int(lines[i, 0]), int(lines[i, 1]))
            pt2 = (int(lines[i, 2]), int(lines[i, 3]))
            width = lines[i, 4]
            cv2.line(mask, pt1, pt2, 0, int(np.ceil(width / 4)))
        if mode == 'overlay':
            image = overlay_heatmap_on_image(image, mask)
        elif mode == 'heatmap':
            image = mask
        else:
            raise NotImplementedError
        return image

    def __call__(self, img):
        """Detects line segments in the provided image.
        Args:
            img (numpy.ndarray): Grayscale image.
        Returns:
            (2D numpy.ndarray): Detected line segments in format:
                [[point1.x, point1.y, point2.x, point2.y, width]]
        """
        assert isinstance(img, np.ndarray)
        assert len(img.shape) < 3
        rows, cols = img.shape
        img = img.reshape(1, rows * cols).tolist()[0]

        lens = len(img)
        img = (ctypes.c_double * lens)(*img)

        with NamedTemporaryFile(prefix='pylsd-', suffix='.ntl.txt', delete=False) as fp:
            fname = fp.name
            fname_bytes = bytes(fp.name) if sys.version_info < (3, 0) else bytes(fp.name, 'utf8')

        self.lsdlib.lsdGet(img, ctypes.c_int(rows), ctypes.c_int(cols), fname_bytes,
                           ctypes.c_double(self.scale), ctypes.c_double(self.sigma_scale),
                           ctypes.c_double(self.ang_th), ctypes.c_double(self.quant), ctypes.c_double(self.eps),
                           ctypes.c_double(self.density_th), ctypes.c_int(self.n_bins), ctypes.c_double(self.max_grad))

        with open(fname, 'r') as fp:
            output = fp.read()
            cnt = output.strip().split(' ')
            count = int(cnt[0])
            dim = int(cnt[1])
            segments = np.array([float(each) for each in cnt[2:]])
            segments = segments.reshape(count, dim)

        os.remove(fname)
        return segments


class BlobDetector():
    def __init__(self, min_circle_area, max_circle_area):
        params = cv2.SimpleBlobDetector_Params()
        params.filterByArea = True
        params.minArea = min_circle_area
        params.maxArea = max_circle_area
        params.filterByCircularity = True
        self.detector = cv2.SimpleBlobDetector_create(params)

    def __call__(self, image):
        keypoints = self.detector.detect(image)
        return keypoints


def check_is_gray(image):
    return len(image.shape) == 2


def overlay_heatmap_on_image(image, heatmap):
    image = np.asarray(255 * (image / (1e-3 + np.amax(image))), np.uint8)
    heatmap = np.asarray(255 * (heatmap / (1e-3 + np.amax(heatmap))), np.uint8)
    if check_is_gray(image):
        image = image[..., None].repeat(3, axis=-1)


    blur = cv2.GaussianBlur(heatmap, (13, 13), 11)
    heatmap_img = cv2.applyColorMap(255-blur, cv2.COLORMAP_JET)


    super_imposed_img = cv2.addWeighted(heatmap_img, 0.5, image, 0.5, 0)
    return super_imposed_img


def binarize_image(image, upper_limit):
    assert np.amax(image) > 1
    image[image < upper_limit] = 0
    image = 255 - image
    return image


def crop_image(image, row_id, column_id, height, width):
    image = image[row_id:row_id + int(height), column_id:column_id + int(width)]
    return image


def rotate_image_around_center_with_theta(image, center, theta):
    '''
    Rotates OpenCV image around center with angle theta (in deg)
    then crops the image according to width and height.
    '''

    # Uncomment for theta in radians
    # theta *= 180/np.pi

    height, width = image.shape[:2]  # cv2.warpAffine expects shape in (length, height)

    matrix = cv2.getRotationMatrix2D(center=center, angle=theta, scale=1)
    image = cv2.warpAffine(src=image, M=matrix, dsize=(width, height))

    return image


def find_contours(img):
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (11, 11))
    morphed = cv2.morphologyEx(img, cv2.MORPH_CLOSE, kernel)
    contours = cv2.findContours(morphed, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    return contours[-2]


def mask_from_contours(ref_img, contours):
    mask = np.zeros(ref_img.shape, np.uint8)
    mask = cv2.drawContours(mask, contours, -1, (1, 1, 1), -1)
    return cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)


def get_foreground_mask(img, margin=0, convex=False):
    foreground = cv2.cvtColor(remove(img), cv2.COLOR_BGR2GRAY)
    mask = np.uint8(foreground > 0)

    if convex:
        cnts = find_contours(mask)
        mask = mask_from_contours(img, cnts)

    if margin > 0:
        mask_resized = mask[margin:-margin, margin:-margin]
        mask_resized = cv2.resize(mask_resized, mask.shape)
        mask = mask_resized

    if mask.shape != img.shape:
        mask = mask[..., None].repeat(3, axis=-1)

    return mask


def concatenate_two_image(img0, img1, axis):
    if len(img0.shape) != len(img1.shape):
        if len(img0.shape) != 3:
            img0 = np.repeat(img0[..., None], repeats=3, axis=-1)
        else:
            img1 = np.repeat(img1[..., None], repeats=3, axis=-1)

    img_cat = np.concatenate([img0, img1], axis=axis)

    return img_cat


def load_image(path, mode):
    if mode == 'BGR':
        return cv2.imread(path)
    elif mode == 'grey':
        return cv2.imread(path, 0)
    elif mode == 'RGB':
        return cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)
    else:
        raise NotImplementedError

def save_image(path, image):
    assert isinstance(path, str)
    cv2.imwrite(path, image)


def show_image(image, title=None):
    matplotlib.use('TkAgg')
    if isinstance(image, list) or (len(image.shape) == 4):
        fig, axs = plt.subplots(1,len(image))
        for image_i, ax in zip(image, axs):
            ax.imshow(image_i)
            ax.axis('off')
    else:
        plt.imshow(image)
        plt.axis('off')

    if title is not None:
        plt.title(title)
    plt.show()
    plt.close()


def align_img(im1, kp1, im2_w=128, im2=None, kp2=None):
    if check_is_gray(im1):
        im1 = im1[..., None].repeat(3, axis=-1)
    if (im2 is not None) and check_is_gray(im2):
        im2 = im2[..., None].repeat(3, axis=-1)

    kp1 = np.float32(kp1)
    kp2 = np.float32([[0, 0], [1, 0], [1, 1], [0, 1]]) * im2_w if kp2 is None else np.float32(kp2)

    im2 = np.zeros((im2_w, im2_w, 3)) if im2 is None else im2


    # Find homography
    h = cv2.getPerspectiveTransform(kp1, kp2)
    # Use homography
    height, width = im2.shape[:2]
    im1Reg = cv2.warpPerspective(im1, h, (width, height))



    if im2 is not None:
        im1Reg[im1Reg==0] = im2[im1Reg==0]
    return im1Reg



import torch
import matplotlib.pyplot as plt
import numpy as np
import os
import seaborn as sns

from collections import defaultdict
import matplotlib.patches as patches


def save_traj(collector, output_for_goal_representation_learning):
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']
    wall = output_for_goal_representation_learning['wall']
    batch, _ = collector.buffer.sample(batch_size=0)
    s_ids = batch.obs_next.state_id
    g_ids = batch.obs_next.goal_id
    ds = batch.done

    traj_infos = defaultdict(dict)
    for s_id, g_id, d_i in zip(s_ids, g_ids, ds):
        s = output_for_goal_representation_learning['state_set'][s_id]
        if g_id not in traj_infos.keys():
            traj_infos[g_id]['goal_state'] = None
            traj_infos[g_id]['state_visitation'] = np.zeros_like(wall)
        traj_infos[g_id]['goal_state'] = goal_id_to_goal_info[g_id]['goal_state']
        traj_infos[g_id]['state_visitation'][s[0], s[1]] += 1

    return traj_infos


def draw_traj(epoch, mode, collector, output_for_goal_representation_learning, test_env, args, enlarge_goal_id=None):
    def draw_training_goal_state(ax):
        for training_goal_id in [elem for elem in goal_id_to_goal_info.keys() if goal_id_to_goal_info[elem]['mode'] == 'train']:
            c = c_dict[goal_id_to_goal_info[training_goal_id]['room_id']]
            marker = {'train': 'o', 'test': '*'}['train']
            g_s = goal_id_to_goal_info[training_goal_id]['goal_state']
            ax.scatter(g_s[1] + 0.5, g_s[0] + 0.5, marker=marker, c=c, s=200)
        return ax

    c_dict = output_for_goal_representation_learning['c_dict']
    state_set = output_for_goal_representation_learning['state_set']
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']
    wall = output_for_goal_representation_learning['wall']

    traj_info = save_traj(collector=collector,
                          output_for_goal_representation_learning=output_for_goal_representation_learning)

    for cur_goal_id in [elem for elem in goal_id_to_goal_info.keys() if goal_id_to_goal_info[elem]['mode'] == mode]:

        if (enlarge_goal_id is not None) and (mode == 'test') and (cur_goal_id!=enlarge_goal_id): continue

        c = c_dict[goal_id_to_goal_info[cur_goal_id]['room_id']]
        marker = {'train': 'o', 'test': '*'}[mode]

        ax = plt.gca()
        ax.axis('off')
        # ax.set_aspect('auto')


        
        heatmap = sns.heatmap(ax=ax, data=traj_info[cur_goal_id]['state_visitation'], cmap='OrRd')
        cbar = heatmap.collections[0].colorbar
        cbar.set_ticks([])
        cbar.set_label('state visitation count for test goal')

        obs = test_env.reset()
        s_id = obs['state_id']
        init_s = state_set[s_id]
        ax.scatter(init_s[0] + 0.5, init_s[1] + 0.5, marker='s', c='b', s=200)
        for goal_id in traj_info.keys():
            if goal_id != cur_goal_id: continue
            g_s = traj_info[goal_id]['goal_state']
            ax.scatter(g_s[1] + 0.5, g_s[0] + 0.5, marker=marker, c=c, s=400)

        for block_pos in np.stack(np.where(wall == 1), axis=-1):
            ax.add_patch(
                patches.Rectangle(xy=(block_pos[1], block_pos[0]), width=1, height=1, facecolor='k', fill=True))

        ax = draw_training_goal_state(ax)

        # for label, (marker, s) in {'training goal': ('o', 200), 'test goal': ('*', 400)}.items():
        #     plt.scatter([], [], s=s, marker=marker, label=label, facecolors='none', edgecolors='k')
        # plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.01), shadow=True, ncol=2)  # , markerscale=3

        plt.savefig(os.path.join(args.logdir, f'{epoch}_{mode}_{args.rl}_{args.method}_{cur_goal_id}.png'),
                    bbox_inches='tight')
        plt.close()

